"use strict";

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

const Middleware = require("../middleware");
const ConnectionManager = require("./connectionManager");
const ObjectID = require("mongodb").ObjectID;

var _require = require("../validators");

const isValidType = _require.isValidType,
      isObject = _require.isObject,
      isEmptyValue = _require.isEmptyValue,
      isString = _require.isString,
      isInEnum = _require.isInEnum,
      isNumber = _require.isNumber;

//const { to } = require('await-to-js');

const defaultTenantsOptions = {
  noListener: false,
  returnNonCachedInstance: false
};

/**
 * DEFAULT METHODS OPTIONS
 */
const defaultFindOptions = {
  moltyClass: true,
  limit: 0,
  sort: null,
  projection: null,
  skip: 0,
  hint: null,
  explain: false,
  timeout: false,
  tailable: false,
  returnKey: false,
  min: null,
  max: null,
  showDiskLoc: false,
  comment: null,
  raw: false,
  promoteLongs: true,
  promoteValues: true,
  promoteBuffers: false,
  readPreference: null,
  partial: false,
  maxTimeMS: null,
  collation: null
};

const defaultInsertOneOptions = {
  moltyClass: true,
  forceServerObjectId: true
};

const defaultInsertManyOptions = {
  moltyClass: true,
  ordered: false,
  forceServerObjectId: true
};

const defaultUpdateOneOptions = {
  moltyClass: true,
  upsert: false
};

const defaultUpdateManyOptions = {
  moltyClass: true,
  upsert: false
};

const defaultAggregateOptions = {
  moltyClass: true
};

const defaultDeleteOneOptions = {
  moltyClass: true
};

const defaultDeleteManyOptions = {
  moltyClass: true
};

class MongoClient {
  constructor() {
    this._connectionManager = null;
    this.models = {};
    this.tenants = {};
    this._indexes = {};
    this._tenantsOptions = {};
  }

  /**
   * connect(): Connect to a database through the DBClient
   *
   * @param {Object} options
   *
   *  @returns {Promise}
   */
  connect(options) {
    this.models = {};
    this.tenants = {};
    this._indexes = {};
    this._tenantsOptions = Object.assign({}, defaultTenantsOptions, options.tenants);

    try {
      this._connectionManager = new ConnectionManager(options);

      return this;
    } catch (error) {
      throw error;
    }
  }

  /**
   * _getIndexesColection():  Get schema indexes and custom indexes
   * we will create on the specific tenant in the DB.
   *
   * @param {Model} model
   */
  _getIndexesColection(model) {
    const schema = model._schemaNormalized;
    let indexes = this._indexes[model._modelName] ? this._indexes[model._modelName] : [];

    Object.keys(schema).forEach(field => {
      if ("unique" in schema[field] && schema[field].unique) {
        if (!indexes.some(index => !!index.key[field])) {
          indexes.push({ key: { [field]: 1 }, unique: true, name: field });
        }
      }
    });

    if (model._schemaOptions && model._schemaOptions.mongoDBIndexes) {
      indexes = indexes.concat(model._schemaOptions.mongoDBIndexes);
    }
    this._indexes[model._modelName] = indexes;
  }

  /**
   * addModel(): Add model to the MongoClient instance so we know
   * all the models side effects needs to take on the DB like
   * ensure indexes
   *
   * @param {Model} model
   */
  addModel(model) {
    if (this._connectionManager === null) throw new Error("You must first connect to the DB before creating a model.");

    if (model._discriminator && this.models[model._discriminator]) {
      throw new Error("There is already a model with the same name: " + model._discriminator);
    } else if (!model._discriminator && this.models[model._modelName]) {
      throw new Error("There is already a model with the same name: " + model._modelName);
    }

    this._getIndexesColection(model);

    if (model._discriminator) {
      this.models[model._discriminator] = model;
    } else {
      this.models[model._modelName] = model;
    }

    // Update models into the connection manager
    this._connectionManager.models = this.models;
  }

  /**
   * _applyTimestamps(): Checkf if the model has associtaed
   * any options from the schema settings like timestamp and apply
   * to the object
   *
   * @param {Object} obj
   * @param {Sting} operation ('insert', 'update')
   */
  _applyTimestamps(obj, model, operation) {
    if (model._schemaOptions.timestamps) {
      if (operation === "insert") {
        obj._data["createdAt"] = new Date();
      }

      if (operation === "update") {
        obj["$currentDate"] = {
          updatedAt: true
        };
      }
    }

    return obj;
  }

  /**
   * ObjectId(): Get a unique ObjectId from MongoDB
   *
   * @returns {ObjectID}
   */
  ObjectId() {
    return new ObjectID();
  }

  /**
   * _applyHooks(): Promisify and binding document or query to the hooks
   *
   * @param [{Object}] hooksList
   * @param {Object} objectBinded
   *
   * @returns [{Object}]
   */
  _applyHooks(hooksList, objectBinded, tenant, meta = {}) {
    let insertHooks = new Middleware();
    let insertManyHooks = new Middleware();
    let updateOneHooks = new Middleware();
    let updateManyHooks = new Middleware();
    let deleteOneHooks = new Middleware();
    let deleteManyHooks = new Middleware();

    if (hooksList.length > 0) {
      hooksList.forEach(key => {
        switch (key.hook) {
          case "insertOne":
            insertHooks.use(key.fn.bind(objectBinded, this, tenant, meta));
            break;
          case "insertMany":
            insertManyHooks.use(key.fn.bind(objectBinded, this, tenant, meta));
            break;
          case "updateOne":
            updateOneHooks.use(key.fn.bind(objectBinded, this, tenant, meta));
            break;
          case "updateMany":
            updateManyHooks.use(key.fn.bind(objectBinded, this, tenant, meta));
            break;
          case "deleteOne":
            deleteOneHooks.use(key.fn.bind(objectBinded, this, tenant, meta));
            break;
          case "deleteMany":
            deleteManyHooks.use(key.fn.bind(objectBinded, this, tenant, meta));
            break;
          default:
            throw new Error('Hook "' + key.hook + '" is not allowed.');
            break;
        }
      });
    }
    return {
      insertOne: insertHooks,
      insertMany: insertManyHooks,
      updateOne: updateOneHooks,
      updateMany: updateManyHooks,
      deleteOne: deleteOneHooks,
      deleteMany: deleteManyHooks
    };
  }

  /**
   * _normalizeAggregatePipeline(): Normalize $looup stages in the aggregate
   * pipeline in case of discriminated models
   *
   * @param {Object} pipeline
   *
   * @returns {Array}
   */
  _normalizeAggregatePipeline(pipeline, parentDiscriminator) {
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = pipeline[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        let stage = _step.value;

        Object.keys(stage).forEach(operator => {
          if (operator === "$lookup") {
            // Check if there is a $lookup operator in the pipeline with
            // discriminated models pointing out

            let _discriminator = null;
            let lookupFrom = stage["$lookup"].from;
            if (this.models[lookupFrom]) {
              _discriminator = this.models[lookupFrom]._discriminator;
              if (_discriminator) {
                stage["$lookup"].from = this.models[lookupFrom]._modelName;
              }
            }

            if (stage["$lookup"]["pipeline"]) {
              return this._normalizeAggregatePipeline(stage["$lookup"]["pipeline"], _discriminator);
            }
          }
        });
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }

    if (parentDiscriminator) {
      const discriminatorKey = this.models[parentDiscriminator]._schemaOptions.inheritOptions.discriminatorKey;


      if (pipeline[0]["$match"] && pipeline[0]["$match"].discriminatorKey) throw new Error('You can not include a specific value for the "discriminatorKey" on the query.');

      const pipelineKeys = Object.keys(pipeline[0]);

      pipeline.unshift({
        $match: { [discriminatorKey]: parentDiscriminator }
      });
    }

    return pipeline;
  }

  /**
   * _getModelCollectionAndDiscriminator(): Form a document it returns
   * model, collection, discriminatorKey and the discriminator
   * @param {Document} doc
   */
  _getModelCollectionAndDiscriminator(doc) {
    let model = {},
        collection = null,
        discriminatorKey = null,
        discriminator = null;

    if (doc._discriminator) {
      // If the documents we are inserting are from a discriminator model
      if (!this.models[doc._discriminator]) throw new Error("The collection " + collection + "does not exist and is not registered.");
      model = this.models[doc._discriminator];
      discriminatorKey = doc._options.inheritOptions.discriminatorKey;
      discriminator = doc._discriminator;
    } else if (this.models[doc._modelName] && Object.keys(this.models[doc._modelName]._childsModels).length > 0) {
      // Else, if the documents are from parent model but has discriminators
      // models
      model = this.models[doc._modelName];
      discriminatorKey = doc._options.inheritOptions.discriminatorKey;
      discriminator = doc._modelName;
    } else {
      // If the docs are not discriminator docs and do not came from a model
      // which has discriminators model
      if (!this.models[doc._modelName]) throw new Error("The collection " + collection + "does not exist and is not registered.");
      model = this.models[doc._modelName];
    }

    collection = doc._modelName;

    return [model, collection, discriminatorKey, discriminator];
  }

  /**
   * insertOne(): Insert one document into the collection and the
   * tenant specifyed. "forceServerObjectId" is set to true in order
   * to require server side assign _id
   *
   * @param {String} tenant
   * @param {Document} doc
   *
   * @returns {Promise}
   */
  insertOne(tenant, doc, options = {}) {
    var _this = this;

    return _asyncToGenerator(function* () {
      const Document = require("../document");
      if (!(doc instanceof Document)) throw new Error("The document should be a proper Document instance");
      if (!tenant && typeof tenant != "string") throw new Error("The document must specify the tenant name (String), got: " + tenant);

      // Assign default options to perform the inserOne query
      const insertOneOptions = Object.assign({}, defaultInsertOneOptions, options);

      const moltyClassEnabled = insertOneOptions.moltyClass;
      delete insertOneOptions.moltyClass;

      // If we are inserting a resources in a discriminator model
      // we have to set the proper params and address to the parent collection

      var _getModelCollectionAn = _this._getModelCollectionAndDiscriminator(doc),
          _getModelCollectionAn2 = _slicedToArray(_getModelCollectionAn, 4);

      const model = _getModelCollectionAn2[0],
            collection = _getModelCollectionAn2[1],
            discriminatorKey = _getModelCollectionAn2[2],
            discriminator = _getModelCollectionAn2[3];

      // If they have discriminator

      if (discriminator && discriminatorKey) {
        if (doc._data[discriminatorKey]) throw new Error('You can not include a specific value for the "discriminatorKey" on the doc.');

        doc._data[discriminatorKey] = discriminator;
      }

      const docAux = _this._applyTimestamps(doc, model, "insert");
      // Ensure index are created
      if (_this._indexes[collection] && _this._indexes[collection].length > 0) {
        yield _this.createIndexes(tenant, collection, _this._indexes[collection]);
      }
      // Apply hooks
      const _preHooksAux = _this._applyHooks(model._preHooks, doc, tenant);
      const _postHooksAux = _this._applyHooks(model._postHooks, doc, tenant);
      // Running pre insert hooks
      yield _preHooksAux.insertOne.exec();
      // Acquiring db instance
      const conn = yield _this._connectionManager.acquire();
      return new Promise((() => {
        var _ref = _asyncToGenerator(function* (resolve, reject) {
          try {
            const result = yield conn.db(tenant, _this._tenantsOptions).collection(collection).insertOne(doc._data, insertOneOptions);

            // Running post insert hooks
            yield _postHooksAux.insertOne.exec();

            if (result) {
              if (moltyClassEnabled) {
                // Binding model properties to the document
                const Document = require("../document");

                const docInserted = new Document(result.ops[0], model._preHooks, model._postHooks, model._methods, model._schemaOptions, model._modelName, model._discriminator);

                resolve(docInserted);
              } else {
                resolve(result);
              }
            } else {
              resolve(result);
            }
          } catch (error) {
            reject(error);
          } finally {
            return yield _this._connectionManager.release(conn);
          }
        });

        return function (_x, _x2) {
          return _ref.apply(this, arguments);
        };
      })());
    })();
  }

  /**
   * insertMany(): Inserts an array of documents into MongoDB.
   * If documents passed in do not contain the _id field,
   * one will be added to each of the documents missing it by
   * the driver, mutating the document. This behavior
   * can be overridden by setting the forceServerObjectId flag.
   *
   * @param {String} tenant
   * @param [{Document}] docs
   *
   * @returns {Promise}
   */
  insertMany(tenant, docs, options = {}) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      if (!(docs instanceof Array)) throw new Error("The documents should be a proper Array instance with documents");
      if (docs.length <= 0) throw new Error("The documents Array should not be empty");
      const Document = require("../document");
      if (!(docs[0] instanceof Document)) throw new Error("Elements of the Array should be Document instances");
      if (!tenant && typeof tenant != "string") throw new Error("The document must specify the tenant name (String), got: " + tenant);

      // Assign default options to perform the inserMany query
      const insertManyOptions = Object.assign({}, defaultInsertManyOptions, options);

      const moltyClassEnabled = insertManyOptions.moltyClass;
      delete insertManyOptions.moltyClass;

      // If we are inserting a resources in a discriminator model
      // we have to set the proper params and address to the parent collection

      var _getModelCollectionAn3 = _this2._getModelCollectionAndDiscriminator(docs[0]),
          _getModelCollectionAn4 = _slicedToArray(_getModelCollectionAn3, 4);

      const model = _getModelCollectionAn4[0],
            collection = _getModelCollectionAn4[1],
            discriminatorKey = _getModelCollectionAn4[2],
            discriminator = _getModelCollectionAn4[3];


      let arrayDocsData = [];

      // If they have discriminator
      if (discriminator && discriminatorKey) {
        // Loop over each document in the Array
        for (let i = 0; i < docs.length; i++) {
          if (docs[i]._discriminator !== discriminator) throw new Error("There are documents in the Array that belongs to different collections.");

          if (docs[i]._data[discriminatorKey]) throw new Error('You can not include a specific value for the "discriminatorKey" on the doc.');

          docs[i]._data[discriminatorKey] = discriminator;

          // Check and apply document options before saving
          const docAux = _this2._applyTimestamps(docs[i], model, "insert");

          // Preparing the doc
          arrayDocsData.push(docAux._data);
        }
      } else {
        // Loop over each document in the Array
        for (let i = 0; i < docs.length; i++) {
          if (docs[i]._modelName !== model._modelName) throw new Error("There are documents in the Array that belongs to different collections.");

          // Check and apply document options before saving
          const docAux = _this2._applyTimestamps(docs[i], model, "insert");

          // Preparing the doc
          arrayDocsData.push(docAux._data);
        }
      }

      // Ensure index are created
      if (_this2._indexes[collection] && _this2._indexes[collection].length > 0) {
        yield _this2.createIndexes(tenant, collection, _this2._indexes[collection]);
      }

      // Apply hooks
      const _preHooksAux = _this2._applyHooks(model._preHooks, docs, tenant);
      const _postHooksAux = _this2._applyHooks(model._postHooks, docs, tenant);

      // Running pre insert hooks
      yield _preHooksAux.insertMany.exec();

      // Acquiring db instance
      const conn = yield _this2._connectionManager.acquire();

      return new Promise((() => {
        var _ref2 = _asyncToGenerator(function* (resolve, reject) {
          try {
            const result = yield conn.db(tenant, _this2._tenantsOptions).collection(collection).insertMany(arrayDocsData, insertManyOptions);

            // Running post insert hooks
            yield _postHooksAux.insertMany.exec();

            if (result) {
              if (moltyClassEnabled) {
                let docInserted = [];
                for (let i = 0; i < result.ops.length; i++) {
                  // Binding model properties to the document
                  const Document = require("../document");

                  docInserted.push(new Document(result.ops[i], model._preHooks, model._postHooks, model._methods, model._schemaOptions, model._modelName, model._discriminator));
                }

                resolve(docInserted);
              } else {
                resolve(result);
              }
            } else {
              resolve(result);
            }
          } catch (error) {
            reject(error);
          } finally {
            return yield _this2._connectionManager.release(conn);
          }
        });

        return function (_x3, _x4) {
          return _ref2.apply(this, arguments);
        };
      })());
    })();
  }

  /**
   * find(): Fetches the first document that matches the query
   * into the collection and the tenant specifyed. "
   *
   * @param {String} tenant
   * @param {Object} query
   * @param {Object} options
   *
   * @returns {Promise}
   */
  find(tenant, collection, query = {}, options = {}) {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      if (!tenant && typeof tenant != "string") throw new Error("Should specify the tenant name (String), got: " + tenant);
      if (!collection && typeof collection != "string") throw new Error("Should specify the collection name (String), got: " + collection);

      // Assign default options to perform the find query
      const findOptions = Object.assign({}, defaultFindOptions, options);

      const moltyClassEnabled = findOptions.moltyClass;
      delete findOptions.moltyClass;

      // If we are looking for resources in a discriminator model
      // we have to set the proper filter and addres to the parent collection
      let model = {};
      if (_this3.models[collection]) {
        const _discriminator = _this3.models[collection]._discriminator;


        if (_discriminator) {
          const discriminatorKey = _this3.models[collection]._schemaOptions.inheritOptions.discriminatorKey;


          if (query[discriminatorKey]) throw new Error('You can not include a specific value for the "discriminatorKey" on the query.');

          query[discriminatorKey] = collection;
        }
        model = _this3.models[collection];
        collection = _this3.models[collection]._modelName;
      } else {
        throw new Error("The collection " + collection + "does not exist and is not registered.");
      }

      // Ensure index are created
      if (_this3._indexes[collection] && _this3._indexes[collection].length > 0) {
        yield _this3.createIndexes(tenant, collection, _this3._indexes[collection]);
      }

      // Acquiring db instance
      const conn = yield _this3._connectionManager.acquire();

      return new Promise((() => {
        var _ref3 = _asyncToGenerator(function* (resolve, reject) {
          try {
            // Find Options
            const limit = findOptions.limit,
                  sort = findOptions.sort,
                  projection = findOptions.projection,
                  skip = findOptions.skip,
                  hint = findOptions.hint,
                  explain = findOptions.explain,
                  timeout = findOptions.timeout,
                  tailable = findOptions.tailable,
                  returnKey = findOptions.returnKey,
                  min = findOptions.min,
                  max = findOptions.max,
                  showDiskLoc = findOptions.showDiskLoc,
                  comment = findOptions.comment,
                  raw = findOptions.raw,
                  promoteLongs = findOptions.promoteLongs,
                  promoteValues = findOptions.promoteValues,
                  promoteBuffers = findOptions.promoteBuffers,
                  readPreference = findOptions.readPreference,
                  partial = findOptions.partial,
                  maxTimeMS = findOptions.maxTimeMS,
                  collation = findOptions.collation;

            // Get and run the Cursor

            const result = yield conn.db(tenant, _this3._tenantsOptions).collection(collection).find(query, _extends({}, findOptions)).toArray();

            // Run the cursor
            // http://mongodb.github.io/node-mongodb-native/3.0/api/Cursor.html
            /*const [error, result] = await to(
              cursor
                .limit(findOptions.limit)
                .project(findOptions.projection)
                .toArray(),
            );*/

            if (result) {
              if (moltyClassEnabled) {
                const Document = require("../document");
                let docs = [];
                result.forEach(function (doc) {
                  docs.push(new Document(doc, model._preHooks, model._postHooks, model._methods, model._schemaOptions, model._modelName, model._discriminator));
                });
                resolve(docs);
              } else {
                resolve(result);
              }
            } else {
              resolve(result);
            }
          } catch (error) {
            reject(error);
          } finally {
            return yield _this3._connectionManager.release(conn);
          }
        });

        return function (_x5, _x6) {
          return _ref3.apply(this, arguments);
        };
      })());
    })();
  }

  /**
   * updateOne(): Update a single document on MongoDB.
   *
   * @param {String} tenant
   * @param {String} collection
   * @param {Object} filter  The Filter used to select the document to update
   * @param {Object} payload The update operations to be applied to the document
   * @param {Object} options Optional settings.
   *
   * @returns {Promise}
   */
  updateOne(tenant, collection, filter, payload, options = {}) {
    var _this4 = this;

    return _asyncToGenerator(function* () {
      if (!tenant && typeof tenant != "string") throw new Error("Should specify the tenant name (String), got: " + tenant);
      if (!collection && typeof collection != "string") throw new Error("Should specify the collection name (String), got: " + collection);
      if (!filter && typeof filter != "object") throw new Error("Should specify the filter options, got: " + filter);
      if (!payload && typeof payload != "string") throw new Error("Should specify the payload object, got: " + payload);

      // Assign default options to perform the updateOne query
      const updateOneOptions = Object.assign({}, defaultUpdateOneOptions, options);

      const moltyClassEnabled = updateOneOptions.moltyClass;
      delete updateOneOptions.moltyClass;

      // If we are updating a resources in a discriminator model
      // we have to set the proper filter and addres to the parent collection
      let model = {};
      if (_this4.models[collection]) {
        const _discriminator = _this4.models[collection]._discriminator;


        if (_discriminator) {
          const discriminatorKey = _this4.models[collection]._schemaOptions.inheritOptions.discriminatorKey;


          if (filter[discriminatorKey]) throw new Error('You can not include a specific value for the "discriminatorKey" on the query.');

          filter[discriminatorKey] = collection;
        }

        model = _this4.models[collection];
        collection = _this4.models[collection]._modelName;
      } else {
        throw new Error("The collection " + collection + "does not exist and is not registered.");
      }

      payload = _this4._applyTimestamps(payload, model, "update");

      // Ensure index are created
      if (_this4._indexes[collection] && _this4._indexes[collection].length > 0) {
        yield _this4.createIndexes(tenant, collection, _this4._indexes[collection]);
      }

      // Apply hooks
      const _preHooksAux = _this4._applyHooks(model._preHooks, payload, tenant, filter);
      const _postHooksAux = _this4._applyHooks(model._postHooks, payload, tenant, filter);

      // Running pre update hooks
      yield _preHooksAux.updateOne.exec();

      // Acquiring db instance
      const conn = yield _this4._connectionManager.acquire();

      return new Promise((() => {
        var _ref4 = _asyncToGenerator(function* (resolve, reject) {
          try {
            const result = yield conn.db(tenant, _this4._tenantsOptions).collection(collection).updateOne(filter, payload, updateOneOptions);

            // Running post update hooks
            yield _postHooksAux.updateOne.exec();

            resolve(result.result);
          } catch (error) {
            reject(error);
          } finally {
            return yield _this4._connectionManager.release(conn);
          }
        });

        return function (_x7, _x8) {
          return _ref4.apply(this, arguments);
        };
      })());
    })();
  }

  /**
   * updateMany(): Update many documents on MongoDB.
   *
   * @param {String} tenant
   * @param {String} collection
   * @param {Object} filter  The Filter used to select the document to update
   * @param {Object} payload The update operations to be applied to the document
   * @param {Object} options Optional settings.
   *
   * @returns {Promise}
   */
  updateMany(tenant, collection, filter, payload, options = {}) {
    var _this5 = this;

    return _asyncToGenerator(function* () {
      if (!tenant && typeof tenant != "string") throw new Error("Should specify the tenant name (String), got: " + tenant);
      if (!collection && typeof collection != "string") throw new Error("Should specify the collection name (String), got: " + collection);
      if (!filter && typeof filter != "object") throw new Error("Should specify the filter options, got: " + filter);
      if (!payload && typeof payload != "string") throw new Error("Should specify the payload object, got: " + payload);

      // Assign default options to perform the updateMany query
      const updateManyOptions = Object.assign({}, defaultUpdateManyOptions, options);

      const moltyClassEnabled = updateManyOptions.moltyClass;
      delete updateManyOptions.moltyClass;

      // If we are updating a resources in a discriminator model
      // we have to set the proper filter and addres to the parent collection
      let model = {};
      if (_this5.models[collection]) {
        const _discriminator = _this5.models[collection]._discriminator;


        if (_discriminator) {
          const discriminatorKey = _this5.models[collection]._schemaOptions.inheritOptions.discriminatorKey;


          if (filter[discriminatorKey]) throw new Error('You can not include a specific value for the "discriminatorKey" on the query.');

          filter[discriminatorKey] = collection;
        }

        model = _this5.models[collection];
        collection = _this5.models[collection]._modelName;
      } else {
        throw new Error("The collection " + collection + "does not exist and is not registered.");
      }

      payload = _this5._applyTimestamps(payload, model, "update");

      // Ensure index are created
      if (_this5._indexes[collection] && _this5._indexes[collection].length > 0) {
        yield _this5.createIndexes(tenant, collection, _this5._indexes[collection]);
      }

      // Apply hooks
      const _preHooksAux = _this5._applyHooks(model._preHooks, payload, tenant, filter);
      const _postHooksAux = _this5._applyHooks(model._postHooks, payload, tenant, filter);

      // Running pre update hooks
      yield _preHooksAux.updateMany.exec();

      // Acquiring db instance
      const conn = yield _this5._connectionManager.acquire();

      return new Promise((() => {
        var _ref5 = _asyncToGenerator(function* (resolve, reject) {
          try {
            const result = yield conn.db(tenant, _this5._tenantsOptions).collection(collection).updateMany(filter, payload, updateManyOptions);

            // Running post update hooks
            yield _postHooksAux.updateMany.exec();

            resolve(result.result);
          } catch (error) {
            reject(error);
          } finally {
            return yield _this5._connectionManager.release(conn);
          }
        });

        return function (_x9, _x10) {
          return _ref5.apply(this, arguments);
        };
      })());
    })();
  }

  /**
   * aggregate(): Execute an aggregation framework pipeline against the collection.
   *
   * @param {String} tenant
   * @param {String} collection
   * @param {Object[]} pipeline Array containing all the aggregation framework commands for the execution
   * @param {Object} options Optional settings.
   *
   * @returns {Promise}
   */
  aggregate(tenant, collection, pipeline = [], options = {}) {
    var _this6 = this;

    return _asyncToGenerator(function* () {
      if (!tenant && typeof tenant != "string") throw new Error("Should specify the tenant name (String), got: " + tenant);
      if (!collection && typeof collection != "string") throw new Error("Should specify the collection name (String), got: " + collection);

      // Assign default options to perform the aggregate query
      const aggregateOptions = Object.assign({}, defaultAggregateOptions, options);

      const moltyClassEnabled = aggregateOptions.moltyClass;
      delete aggregateOptions.moltyClass;

      // If we are looking for resources in a discriminator model
      // we have to set the proper filter and address to the parent collection
      let model = {},
          discriminator = null;
      if (_this6.models[collection]) {
        const _discriminator = _this6.models[collection]._discriminator;


        if (_discriminator) {
          const discriminatorKey = _this6.models[collection]._schemaOptions.inheritOptions.discriminatorKey;


          if (pipeline[0]["$match"] && pipeline[0]["$match"].discriminatorKey) throw new Error('You can not include a specific value for the "discriminatorKey" on the query.');

          discriminator = collection;
        }

        model = _this6.models[collection];
        collection = _this6.models[collection]._modelName;
      } else {
        throw new Error("The collection " + collection + "does not exist and is not registered.");
      }

      // Check all stages in the pipeline looking for any lookup stage
      // pointing out to a discriminated model
      pipeline = _this6._normalizeAggregatePipeline(pipeline, discriminator);

      // Ensure index are created
      if (_this6._indexes[collection] && _this6._indexes[collection].length > 0) {
        yield _this6.createIndexes(tenant, collection, _this6._indexes[collection]);
      }

      // Acquiring db instance
      const conn = yield _this6._connectionManager.acquire();

      return new Promise((() => {
        var _ref6 = _asyncToGenerator(function* (resolve, reject) {
          try {
            // Get and run the Cursor
            const result = yield conn.db(tenant, _this6._tenantsOptions).collection(collection).aggregate(pipeline, aggregateOptions).toArray();

            // Run the cursor
            // http://mongodb.github.io/node-mongodb-native/3.0/api/Cursor.html
            /*const [error, result] = await to(
              cursor
                .limit(findOptions.limit)
                .project(findOptions.projection)
                .toArray(),
            );*/

            if (result) {
              if (moltyClassEnabled) {
                /* const Document = require('../document');
                  let docs = [];
                  result.forEach(doc => {
                    docs.push(
                      new Document(
                        doc,
                        model._preHooks,
                        model._postHooks,
                        model._methods,
                        model._schemaOptions,
                        model._modelName,
                        model._discriminator,
                      ),
                    );
                  });
                  resolve(docs);*/
                resolve(result);
              } else {
                resolve(result);
              }
            } else {
              resolve(result);
            }
          } catch (error) {
            reject(error);
          } finally {
            return yield _this6._connectionManager.release(conn);
          }
        });

        return function (_x11, _x12) {
          return _ref6.apply(this, arguments);
        };
      })());
    })();
  }

  /**
   * deleteOne(): Delete a single document on MongoDB.
   *
   * @param {String} tenant
   * @param {String} collection
   * @param {Object} filter  The Filter used to select the document to delete
   * @param {Object} options Optional settings.
   *
   * @returns {Promise}
   */
  deleteOne(tenant, collection, filter, options = {}) {
    var _this7 = this;

    return _asyncToGenerator(function* () {
      if (!tenant && typeof tenant != "string") throw new Error("Should specify the tenant name (String), got: " + tenant);
      if (!collection && typeof collection != "string") throw new Error("Should specify the collection name (String), got: " + collection);
      if (!filter && typeof filter != "object") throw new Error("Should specify the filter options, got: " + filter);

      // Assign default options to perform the deleteOne query
      const deleteOneOptions = Object.assign({}, defaultDeleteOneOptions, options);

      const moltyClassEnabled = deleteOneOptions.moltyClass;
      delete deleteOneOptions.moltyClass;

      // If we are deleting a resources in a discriminator model
      // we have to set the proper filter and addres to the parent collection
      let model = {};
      if (_this7.models[collection]) {
        const _discriminator = _this7.models[collection]._discriminator;


        if (_discriminator) {
          const discriminatorKey = _this7.models[collection]._schemaOptions.inheritOptions.discriminatorKey;


          if (filter[discriminatorKey]) throw new Error('You can not include a specific value for the "discriminatorKey" on the query.');

          filter[discriminatorKey] = collection;
        }

        model = _this7.models[collection];
        collection = _this7.models[collection]._modelName;
      } else {
        throw new Error("The collection " + collection + "does not exist and is not registered.");
      }

      // Ensure index are created
      if (_this7._indexes[collection] && _this7._indexes[collection].length > 0) {
        yield _this7.createIndexes(tenant, collection, _this7._indexes[collection]);
      }

      // Apply hooks
      const _preHooksAux = _this7._applyHooks(model._preHooks, {}, tenant, filter);
      const _postHooksAux = _this7._applyHooks(model._postHooks, {}, tenant, filter);

      // Running pre delete hooks
      yield _preHooksAux.deleteOne.exec();

      // Acquiring db instance
      const conn = yield _this7._connectionManager.acquire();

      return new Promise((() => {
        var _ref7 = _asyncToGenerator(function* (resolve, reject) {
          try {
            const result = yield conn.db(tenant, _this7._tenantsOptions).collection(collection).deleteOne(filter, deleteOneOptions);

            // Running post delete hooks
            yield _postHooksAux.deleteOne.exec();

            resolve(result.result);
          } catch (error) {
            reject(error);
          } finally {
            return yield _this7._connectionManager.release(conn);
          }
        });

        return function (_x13, _x14) {
          return _ref7.apply(this, arguments);
        };
      })());
    })();
  }

  /**
   * deleteMany(): Delete many documents on MongoDB.
   *
   * @param {String} tenant
   * @param {String} collection
   * @param {Object} filter  The Filter used to select the document to delete
   * @param {Object} options Optional settings.
   *
   * @returns {Promise}
   */
  deleteMany(tenant, collection, filter, options = {}) {
    var _this8 = this;

    return _asyncToGenerator(function* () {
      if (!tenant && typeof tenant != "string") throw new Error("Should specify the tenant name (String), got: " + tenant);
      if (!collection && typeof collection != "string") throw new Error("Should specify the collection name (String), got: " + collection);
      if (!filter && typeof filter != "object") throw new Error("Should specify the filter options, got: " + filter);

      // Assign default options to perform the deleteOne query
      const deleteManyOptions = Object.assign({}, defaultDeleteManyOptions, options);

      const moltyClassEnabled = deleteManyOptions.moltyClass;
      delete deleteManyOptions.moltyClass;

      // If we are deleting a resources in a discriminator model
      // we have to set the proper filter and addres to the parent collection
      let model = {};
      if (_this8.models[collection]) {
        const _discriminator = _this8.models[collection]._discriminator;


        if (_discriminator) {
          const discriminatorKey = _this8.models[collection]._schemaOptions.inheritOptions.discriminatorKey;


          if (filter[discriminatorKey]) throw new Error('You can not include a specific value for the "discriminatorKey" on the query.');

          filter[discriminatorKey] = collection;
        }

        model = _this8.models[collection];
        collection = _this8.models[collection]._modelName;
      } else {
        throw new Error("The collection " + collection + "does not exist and is not registered.");
      }

      // Ensure index are created
      if (_this8._indexes[collection] && _this8._indexes[collection].length > 0) {
        yield _this8.createIndexes(tenant, collection, _this8._indexes[collection]);
      }

      // Apply hooks
      const _preHooksAux = _this8._applyHooks(model._preHooks, {}, tenant, filter);
      const _postHooksAux = _this8._applyHooks(model._postHooks, {}, tenant, filter);

      // Running pre delete hooks
      yield _preHooksAux.deleteMany.exec();

      // Acquiring db instance
      const conn = yield _this8._connectionManager.acquire();

      return new Promise((() => {
        var _ref8 = _asyncToGenerator(function* (resolve, reject) {
          try {
            const result = yield conn.db(tenant, _this8._tenantsOptions).collection(collection).deleteMany(filter, deleteManyOptions);

            // Running post delete hooks
            yield _postHooksAux.deleteMany.exec();

            resolve(result.result);
          } catch (error) {
            reject(error);
          } finally {
            return yield _this8._connectionManager.release(conn);
          }
        });

        return function (_x15, _x16) {
          return _ref8.apply(this, arguments);
        };
      })());
    })();
  }

  /**
   * createIndexes(): Creates indexes on the db and collection collection
   *
   * @param {String} tenant
   * @param {String} collection
   * @param [{String}] fields
   *
   * @returns {Promise}
   */
  createIndexes(tenant, collection, fields) {
    var _this9 = this;

    return _asyncToGenerator(function* () {
      if (!tenant && typeof tenant != "string") throw new Error("Should specify the tenant name (String), got: " + tenant);
      if (!collection && typeof collection != "string") throw new Error("Should specify the collection name (String), got: " + collection);
      if (!fields && typeof fields != "object") throw new Error("Should specify the field name (Array), got: " + fields);

      // Checking if indexes are already set for this tenant and this collection
      if (_this9.tenants[tenant] && _this9.tenants[tenant][collection]) return;

      // Acquiring db instance
      const conn = yield _this9._connectionManager.acquire();

      // Checking if indexes already exist in the DB
      const indexes = yield conn.db(tenant, _this9._tenantsOptions).collection(collection).indexes();

      const fieldsFiltered = [];
      let indexAlreadyExist = false;
      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;

      try {
        for (var _iterator2 = fields[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          const field = _step2.value;

          const keys = Object.keys(field.key);
          var _iteratorNormalCompletion3 = true;
          var _didIteratorError3 = false;
          var _iteratorError3 = undefined;

          try {
            for (var _iterator3 = indexes[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
              const index = _step3.value;

              const indexKeys = index.weights ? Object.keys(index.weights) : Object.keys(index.key);
              if (keys.every(function (key) {
                return indexKeys.includes(key);
              })) {
                indexAlreadyExist = true;
                break;
              }
            }
          } catch (err) {
            _didIteratorError3 = true;
            _iteratorError3 = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion3 && _iterator3.return) {
                _iterator3.return();
              }
            } finally {
              if (_didIteratorError3) {
                throw _iteratorError3;
              }
            }
          }

          if (!indexAlreadyExist) {
            fieldsFiltered.push(field);
          }

          indexAlreadyExist = false;
        }
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2.return) {
            _iterator2.return();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }

      if (fieldsFiltered.length <= 0) {
        return;
      }

      return new Promise((() => {
        var _ref9 = _asyncToGenerator(function* (resolve, reject) {
          try {
            const result = yield conn.db(tenant, _this9._tenantsOptions).collection(collection).createIndexes(fieldsFiltered);

            // Set Indexes for this tenant and this collection are already set
            _this9.tenants[tenant] = _extends({}, _this9.tenants[tenant], {
              [collection]: true
            });

            resolve(result);
          } catch (error) {
            reject(error);
          } finally {
            return yield _this9._connectionManager.release(conn);
          }
        });

        return function (_x17, _x18) {
          return _ref9.apply(this, arguments);
        };
      })());
    })();
  }

  /**
   * dropDatabase(): Drop a database, removing it permanently from the server.
   *
   * @param {String} tenant
   *
   * @returns {Promise}
   */
  dropDatabase(tenant) {
    var _this10 = this;

    return _asyncToGenerator(function* () {
      if (!tenant && typeof tenant != "string") throw new Error("Should specify the tenant name (String), got: " + tenant);

      // Acquiring db instance
      const conn = yield _this10._connectionManager.acquire();

      return new Promise((() => {
        var _ref10 = _asyncToGenerator(function* (resolve, reject) {
          try {
            const result = yield conn.db(tenant, _this10._tenantsOptions).dropDatabase();

            resolve(result);
          } catch (error) {
            reject(error);
          } finally {
            return yield _this10._connectionManager.release(conn);
          }
        });

        return function (_x19, _x20) {
          return _ref10.apply(this, arguments);
        };
      })());
    })();
  }

  /**
   * dropCollection(): Drop a collection, removing it permanently from the server.
   *
   * @param {String} collection
   * @param {String} tenant
   *
   * @returns {Promise}
   */
  dropCollection(collection, tenant) {
    var _this11 = this;

    return _asyncToGenerator(function* () {
      if (!tenant && typeof tenant != "string") throw new Error("Should specify the tenant name (String), got: " + tenant);

      if (!collection && typeof collection != "string") throw new Error("Should specify the collection name (String), got: " + collection);

      // Acquiring db instance
      const conn = yield _this11._connectionManager.acquire();

      return new Promise((() => {
        var _ref11 = _asyncToGenerator(function* (resolve, reject) {
          try {
            const result = yield conn.db(tenant, _this11._tenantsOptions).dropCollection(collection);

            resolve(result);
          } catch (error) {
            reject(error);
          } finally {
            return yield _this11._connectionManager.release(conn);
          }
        });

        return function (_x21, _x22) {
          return _ref11.apply(this, arguments);
        };
      })());
    })();
  }

  /**
   * executeDbAdminCommand(): Runs a command on the database as admin.
   *
   * @param {Object} command
   * @param {Object} options
   *
   * @returns {Promise}
   */
  executeDbAdminCommand(command, tenant = null, options = {}) {
    var _this12 = this;

    return _asyncToGenerator(function* () {
      if (Object.keys(command).length <= 0) throw new Error("command field should not be empty");

      // Acquiring db instance
      const conn = yield _this12._connectionManager.acquire();

      return new Promise((() => {
        var _ref12 = _asyncToGenerator(function* (resolve, reject) {
          try {
            const result = yield conn.db(tenant, _this12._tenantsOptions).executeDbAdminCommand(command, options);

            resolve(result);
          } catch (error) {
            reject(error);
          } finally {
            return yield _this12._connectionManager.release(conn);
          }
        });

        return function (_x23, _x24) {
          return _ref12.apply(this, arguments);
        };
      })());
    })();
  }
}

module.exports = new MongoClient();