'use strict';

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

const mongoClient = require('./clients/mongoClient');
const Document = require('./document');

var _require = require('await-to-js');

const to = _require.to;

var _require2 = require('./validators');

const isValidType = _require2.isValidType,
      isObject = _require2.isObject,
      isArray = _require2.isArray,
      isEmptyValue = _require2.isEmptyValue,
      isString = _require2.isString,
      isInEnum = _require2.isInEnum,
      isNumber = _require2.isNumber;


class Model {
  /**
   * Model(): Model conmstructor
   *
   * @param {Schema} schema
   * @param {String} modelName
   *
   * @returns {Model}
   */
  constructor(schema, modelName, discriminator = null) {
    if (arguments.length < 1) throw new Error('Schema is missing');

    this._schemaNormalized = schema._schema;
    this._schemaOptions = schema._options;

    this._modelName = modelName.replace(/\s/g, '');

    this._preHooks = schema._preHooks;
    this._postHooks = schema._postHooks;

    this._childsModels = {};

    this._methods = {};
    this._validateStaticMethodsName(schema.methods, this._schemaNormalized);

    this._methods = schema.methods;

    this._discriminator = discriminator;

    this._normalizedPayload = null;

    mongoClient.addModel(this);
  }

  /**
   * _validateStaticMethodsName(): Check if there is an static method
   * with the same name as schema field
   *
   * @param {Object} methods
   * @params {Object} schema
   */
  _validateStaticMethodsName(methods, schema) {
    Object.keys(methods).forEach(key => {
      if (key === '_preHooks' || key === '_postHooks' || key === '_data') throw new Error('Static methods can not be called like: ' + key);

      if (key in schema) throw new Error('Static methods can not be called like an schema field name: ' + key);

      // Objects nested
      if (schema[key] && !schema[key].type && isObject(schema[key])) {
        return this._validateStaticMethodsName(methods, schema[key]);
      }
    });
  }

  /**
   * _validateDiscriminatorName(): Check if there is a schema field
   * with the same name as the discriminator key
   *
   * @param {Object} discriminator
   * @params {Object} schema
   */
  _validateDiscriminatorName(discriminator, schema) {
    Object.keys(schema).forEach(key => {
      if (key === '_preHooks' || key === '_postHooks' || key === '_data') throw new Error('Discriminator name can not be called like this: ' + key);

      if (key === discriminator) throw new Error('Discriminator name can not be called like anothe schema field: ' + key);

      // Objects nested
      if (!schema[key].type && isObject(schema[key])) {
        return this._validateDiscriminatorName(discriminator, schema[key]);
      }
    });
  }

  /**
   * new(): Create a new document based on this model
   *
   * @param {Object} payload
   *
   * @returns {Onject} doc
   */
  new(payload) {
    var _this = this;

    return _asyncToGenerator(function* () {
      if (!payload || Object.keys(payload).length === 0 && payload.constructor === Object) throw new Error('Document payload is required');

      // Check if paylaod field names are correct
      _this._validatePayloadFieldNames(payload, _this._schemaNormalized);

      let data = null;
      // Normalize the payload with the model schema
      data = _this._normalizePayload(payload, _this._schemaNormalized);

      try {
        // Validate all the values
        yield _this.validatePayloadFieldValues(data, _this._schemaNormalized, data);

        // Returning the new document created
        const newDoc = new Document(data, _this._preHooks, _this._postHooks, _this._methods, _this._schemaOptions, _this._modelName, _this._discriminator);

        return newDoc;
      } catch (err) {
        throw err;
      }
    })();
  }

  /**
   * discriminator(): Create a new discriminator for an schema
   * which inherit all the properties from the parent schema
   *
   * @param {Schema} schemaDiscriminator
   * @param {String} discriminatorModelName
   */
  discriminator(schemaDiscriminator, discriminatorModelName) {
    const schemaAux = {};

    schemaAux._options = Object.assign({}, schemaDiscriminator._options);

    var _schemaAux$_options$i = schemaAux._options.inheritOptions;
    const merge = _schemaAux$_options$i.merge,
          childDiscriminatorKey = _schemaAux$_options$i.discriminatorKey;


    if (!childDiscriminatorKey) throw new Error('To create an inherited model you have to sepcify the "discriminatorKey" in the schema "options" field.');

    const parentDiscriminatorKey = this._schemaOptions.inheritOptions.discriminatorKey;


    if (!parentDiscriminatorKey) throw new Error('You can not inherit from this model, you should add the "discriminatorKey" option in both schemas.');

    if (parentDiscriminatorKey !== childDiscriminatorKey) throw new Error('The "discriminatorKey" from the parent and the child model shoul be the same: [' + parentDiscriminatorKey + ',' + childDiscriminatorKey + ']');

    schemaAux.methods = Object.assign({}, schemaDiscriminator.methods);
    if (merge && merge.indexOf('methods') >= 0) {
      Object.keys(schemaDiscriminator.methods).forEach(key => {
        if (key in this._methods) throw new Error('Static methods from a child schema can not be called like another method from the parent: ' + key);
      });

      schemaAux.methods = Object.assign({}, schemaAux.methods, this._methods);
    }

    schemaAux._preHooks = schemaDiscriminator._preHooks;
    if (merge && merge.indexOf('preHooks') >= 0) {
      schemaAux._preHooks = this._preHooks.concat(schemaAux._preHooks);
    }

    schemaAux._postHooks = schemaDiscriminator._postHooks;
    if (merge && merge.indexOf('postHooks') >= 0) {
      schemaAux._postHooks = this._postHooks.concat(schemaAux._postHooks);
    }

    if (this._schemaOptions) {
      // Merge timestamps
      if (this._schemaOptions.timestamps) {
        schemaAux._options.timestamps = true;
      }

      // Merge elasticSearchIndexes
      if (this._schemaOptions.elasticSearchIndexes) {
        schemaAux._options.elasticSearchIndexes = Object.assign({}, schemaAux._options.elasticSearchIndexes, this._schemaOptions.elasticSearchIndexes);
      }
    }

    schemaAux._schema = Object.assign({}, schemaDiscriminator._schema, this._schemaNormalized);

    this._validateDiscriminatorName(childDiscriminatorKey, schemaAux._schema);

    const discriminatorModel = new Model(schemaAux, this._modelName, discriminatorModelName);

    if (this._childsModels[discriminatorModelName]) throw new Error('There is already model discriminator with the same name: ' + discriminatorModelName);

    this._childsModels[discriminatorModelName] = discriminatorModel;

    return discriminatorModel;
  }

  /**
   * _validatePayloadFieldNames(): Check if all the field names
   * from the payload corresponds to schema with.
   *
   * @param {Object} payload
   * @param {schema} schema
   */
  _validatePayloadFieldNames(payload, schema) {
    Object.keys(payload).forEach(key => {
      if (schema[key] === undefined && key !== '_id') {
        throw new Error('Field name ' + key + ' does not correspond to any field name on the schema');
      }
      // Array
      if (isArray(schema[key]) && !schema[key][0].type && payload[key]) {
        payload[key].forEach(arrayItem => {
          return this._validatePayloadFieldNames(arrayItem, schema[key][0]);
        });
      }

      // Objects nested
      if (isObject(schema[key]) && !isArray(schema[key]) && !schema[key].type && isObject(payload[key]) && Object.keys(payload[key]).length > 0) {
        return this._validatePayloadFieldNames(payload[key], schema[key]);
      }
    });
  }

  /**
   * validatePayloadFieldValues : Check if the data type and format is
   * accepted by the schema assigned to this model and
   * also pass all the validation
   *
   * @param {Object} payload
   * @param {Schema} schema
   * @param {String} tenant
   */
  validatePayloadFieldValues(payload = {}, schema, parentPayload) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = Object.keys(schema)[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          let key = _step.value;

          // Array
          if (!schema[key].type && isArray(schema[key]) && payload[key]) {
            var _iteratorNormalCompletion2 = true;
            var _didIteratorError2 = false;
            var _iteratorError2 = undefined;

            try {
              for (var _iterator2 = payload[key][Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                let arrayItem = _step2.value;

                try {
                  yield _this2.validatePayloadFieldValues(arrayItem, schema[key][0], parentPayload);
                  continue;
                } catch (error) {
                  throw error;
                }
              }
            } catch (err) {
              _didIteratorError2 = true;
              _iteratorError2 = err;
            } finally {
              try {
                if (!_iteratorNormalCompletion2 && _iterator2.return) {
                  _iterator2.return();
                }
              } finally {
                if (_didIteratorError2) {
                  throw _iteratorError2;
                }
              }
            }
          }

          // Objects nested
          if (!schema[key].type && isObject(schema[key]) && !isArray(schema[key])) {
            try {
              yield _this2.validatePayloadFieldValues(payload[key], schema[key], parentPayload);
              continue;
            } catch (error) {
              throw error;
            }
          }

          // No required values
          if (!payload || payload[key] === undefined || isEmptyValue(payload[key])) {
            if (!schema[key].required) continue;else throw new Error('Key "' + key + '" is required');
          }

          // Validation type
          if (!isValidType(payload[key], isArray(schema[key]) ? Array : schema[key].type)) {
            throw new Error('Unsuported value (' + JSON.stringify(payload[key]) + ') for type ' + schema[key].type);
          }

          // Reg exp validation
          if (schema[key].match && isString(payload[key]) && !schema[key].match.test(payload[key])) {
            throw new Error('Value assigned to "' + key + '" does not match the regex/string ' + schema[key].match.toString() + '. Value was ' + payload[key]);
          }

          // Enum validation
          if (!isInEnum(schema[key].enum, payload[key])) {
            throw new Error('Value assigned to "' + key + '" should be in enum [' + schema[key].enum.join(', ') + '], got ' + payload[key]);
          }

          // Min value validation
          if (isNumber(schema[key].min) && payload[key] < schema[key].min) {
            throw new Error('Value assigned to "' + key + '" is less than min, ' + schema[key].min + ', got ' + payload[key]);
          }

          // Max value validation
          if (isNumber(schema[key].max) && payload[key] > schema[key].max) {
            throw new Error('Value assigned to "' + key + '" is less than max, ' + schema[key].max + ', got ' + payload[key]);
          }

          // Max lenght validation
          if (schema[key].maxlength && isString(payload[key]) && payload[key].length > schema[key].maxlength) {
            throw new Error('Value assigned to "' + key + '" is bigger than ' + schema[key].maxlength.toString() + '. Value was ' + payload[key].length);
          }

          // Custom validation
          if (typeof schema[key].validate === 'function') {
            let error, isValid;
            if (schema[key].validate.constructor.name === 'AsyncFunction') {
              var _ref = yield to(schema[key].validate(parentPayload));

              var _ref2 = _slicedToArray(_ref, 2);

              error = _ref2[0];
              isValid = _ref2[1];
            } else {
              isValid = schema[key].validate(parentPayload);
            }

            if (!isValid || error) {
              let err = new Error('Value assigned to "' + key + '" failed custom validator. Value was ' + payload[key]);
              err.code = 'VALIDATION_FAILED';
              throw err;
            }
          }
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    })();
  }

  /**
   * _normalizePayload(): Assign default values to the payload
   *
   *
   * @param {Object} payload
   * @param {schema} schema
   *
   * returns {Object} doc
   */
  _normalizePayload(payload, schema) {
    Object.keys(schema).forEach(key => {
      // Ref values (Ids)
      if (schema[key].ref) {
        if (!payload || isEmptyValue(payload[key])) {
          const refNormalValue = isArray(schema[key].type) ? [] : null;
          payload = _extends({}, payload, {
            [key]: refNormalValue
          });
        } else if (payload[key]) {
          const Schema = require('./schema');
          if (isArray(schema[key].type)) {
            for (let i = 0; i < payload[key].length; i += 1) {
              payload[key][i] = Schema.types().ObjectId(payload[key][i]);
            }
          } else {
            payload[key] = Schema.types().ObjectId(payload[key]);
          }
        }
        return;
      }

      // Default values
      if ((!payload || payload[key] === undefined) && 'default' in schema[key]) {
        const defaultValue = typeof schema[key].default === 'function' ? schema[key].default() : schema[key].default;

        if (!payload) payload = {
          [key]: defaultValue
        };else payload[key] = defaultValue;
        return;
      }

      // Date type as string should be converted to date type
      if (schema[key].type === Date && payload && !isEmptyValue(payload[key]) && typeof payload[key] === typeof 'String') {
        payload[key] = new Date(payload[key]);
        return;
      }

      // Array
      if (!schema[key].type && isArray(schema[key]) && payload[key] && payload[key].length > 0) {
        for (let i = 0; i < payload[key].length; i++) {
          payload[key][i] = this._normalizePayload(payload[key][i], schema[key][0]);
        }
        return;
      }

      // Objects nested
      if (!schema[key].type && isObject(schema[key]) && !isArray(schema[key])) {
        const aux = this._normalizePayload(payload ? payload[key] : {}, schema[key]);
        if (aux) payload[key] = aux;
        return;
      }

      // No required values
      if ((!payload || payload[key] === undefined) && !schema[key].required) return;
    });

    return payload;
  }
}

module.exports = Model;